#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2020
# File   : zroll.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 22.02.2021

import typing

import discord
from discord.ext import commands

from .database import fetch_guild_db
from .logging import LoggingMixin
from .rolls import (RollHandler, QuietRollHandler, SekretRollHandler,
                    RollList, DiceDelta)
from .utils import handle_http_exception


class RollCommands(commands.Cog, LoggingMixin):

    def __init__(self, bot, db):
        self.bot = bot
        self.db = db

        super().__init__()

    @commands.command(name='z', help='Evaluate a dice roll.')
    @fetch_guild_db
    @handle_http_exception
    async def zardoz_roll(self, ctx, *, args):

        try:
            roll = RollHandler(ctx, self.log, ctx.variables, args,
                               game_mode=ctx.game_mode)
        except ValueError as e:
            self.log.error(f'Roll handling failed: {e}')
            await ctx.message.reply(f'You fucked up your roll, {ctx.author}. {e}')
        else:
            await roll.add_to_db(ctx.guild_db)
            await ctx.message.reply(roll.msg())

    @commands.command(name='zq', help='Evaluate a dice roll, quietly.')
    @fetch_guild_db
    @handle_http_exception
    async def zardoz_quiet_roll(self, ctx, *, args):

        try:
            roll = QuietRollHandler(ctx, self.log, ctx.variables, args,
                                    game_mode=ctx.game_mode)
        except ValueError as e:
            self.log.error(f'Roll handling failed: {e}')
            await ctx.message.reply(f'You fucked up your roll, {ctx.author}. {e}')
        else:
            await roll.add_to_db(ctx.guild_db)
            await ctx.message.reply(roll.msg())

    @commands.command(name='zs', help='Make a secret roll and DM to member.')
    @fetch_guild_db
    @handle_http_exception
    async def zardoz_secret_roll(self, ctx, member: typing.Optional[discord.Member], *, args):
        if member is None:
            member = ctx.author

        try:
            roll = SekretRollHandler(ctx, self.log, ctx.variables, args,
                                     game_mode=ctx.game_mode, require_tag=True)
        except ValueError as e:
            self.log.error(f'Roll handling failed: {e}')
            await ctx.author.send(f'You fucked up your roll, {ctx.author}. {e}')
        else:
            await roll.add_to_db(ctx.guild_db)
            await member.send(roll.msg())

    @commands.command(name='zr', help='Reroll previous roll')
    @fetch_guild_db
    @handle_http_exception
    async def zroll_reroll(self, ctx, member: typing.Optional[discord.Member]):
        if member is None:
            member = ctx.author

        saved = await ctx.guild_db.get_last_user_roll(member.id)

        if saved is None:
            await ctx.message.reply(f'Ope, no roll history for {member}.')
        else:
            cmd = saved['roll']
            roll = RollHandler(ctx, self.log, ctx.variables, cmd,
                               game_mode=ctx.game_mode)
            await roll.add_to_db(ctx.guild_db)
            await ctx.message.reply(f'Reroll {roll.msg()}')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2021
# File   : rolls.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 22.02.2021

import dice
from dice import DiceBaseException
from discord.ext import commands

import logging
import re

from .database import ZardozDatabase
from .state import GameMode, MODE_DICE
from .utils import SUCCESS, FAILURE


log = logging.getLogger()

OPS = ['+', '-', '<=', '<', '>=', '>', '(', ')', r'#']
DELIMS = [r'\s']
SPLIT_PAT = '|'.join(list(map(re.escape, OPS)) + DELIMS)


def split_tokens(word):
    log.info(f'SPLIT "{word}" on: "{SPLIT_PAT}"')
    return [t for t in re.split(f'({SPLIT_PAT})', word)]


def filter_tokens(tokens):
    return [t for t in tokens if t and not re.match(r'\s', t)]


def tokenize_roll(cmd: str):
    """tokenize_roll.

    Parameters
    ----------
    cmd : str
        Raw command string to tokenize.
    """

    raw = []

    log.info(f'Tokenize: "{cmd}"')
    if isinstance(cmd, str):
        raw = split_tokens(cmd)
    else:
        for arg in cmd:
            raw.extend(split_tokens(arg))

    raw = filter_tokens(raw)

    tokens, tag = [], ''
    for i, token in enumerate(raw):
        if token.startswith('#'):
            tokens = raw[:i]
            tag = ' '.join(raw[i:])
            break
    if not tag:
        tokens = raw
    tag = tag.strip('# ')

    return tokens, tag


def expand_tokens(tokens, mode = None, variables = {}):
    
    if mode is None:
        mode = GameMode.DEFAULT

    if mode is not None and not isinstance(mode, GameMode):
        raise ValueError('mode must be of type GameMode')

    expanded = []

    for token in tokens:
        if token.isnumeric():
            expanded.append(token)
        elif token in OPS:
            expanded.append(token)
        elif token.startswith('$'):
            var = token.strip('$')
            if var in variables:
                val = variables[var]
                if not isinstance(val, int):
                    raise ValueError('Non-integer variable!')
                expanded.append(variables[var])
            else:
                log.error(f'Undefined var: {var}')
                raise ValueError(f'Sorry bruh, that variable isn\'t defined: `{var}`')
        else:
            if token == 'r':
                token = MODE_DICE[mode]

            ndice = re.match(r'^\d+', token)
            if ndice is not None and int(ndice.group(0)) > 69:
                log.error(f'Too many dice.')
                raise ValueError(f'Woah there cowboy, you trying to crash me? Try using 69 or less dice.')
            try:
                resolved = dice.roll(token)
            except DiceBaseException:
                log.error(f'Got invalid token: {token}.')
                raise ValueError(f'I don\'t like this argument: `{token}`.')
            else:
                log.info(f'Rolled a token: {token} -> {resolved}')
                expanded.append(RollList(token, resolved))

    return expanded


def evaluate_expr(expanded_tokens):
    eval_expr = ' '.join([(repr(t) if not isinstance(t, str) else t) for t in expanded_tokens])
    try:
        return eval(eval_expr), eval_expr
    except SyntaxError as e:
        return None, eval_expr


class RollHandler:

    def __init__(self, ctx, log, variables, roll,
                       require_tag=False, game_mode=GameMode.DEFAULT):

        log.info(f'Roll request: {roll}')

        self.ctx = ctx
        self.log = log
        self.game_mode = game_mode
        self.roll = roll

        self.tokens, self.tag = tokenize_roll(roll)
        if require_tag and not self.tag:
            raise ValueError('Add a tag, it\'s the polite thing to do.')
        self.log.info(f'Raw tokens: {self.tokens}')

        self.expanded = expand_tokens(self.tokens,
                                      mode = self.game_mode,
                                      variables = variables)
        self.log.info(f'Expanded tokens: {self.expanded}')
        self.expr =  ' '.join((str(token) for token in self.expanded))
        
        self.result, eval_expr = evaluate_expr(self.expanded)
        if self.result is None:
            log.error(f'Python parsing error: {eval_expr}.')
            raise ValueError(f'Your syntax was scintillating, but I couldn\'t parse it.')

        log.info(f'Handled roll: {str(self)}')

    async def add_to_db(self, db: ZardozDatabase):
        await db.add_roll(self.ctx.author.id, self.ctx.author.nick, self.ctx.author.name,
                          ' '.join(self.tokens), self.tag, self.expr)

    def msg(self):
        header = f':game_die: {self.ctx.author.mention}' + (f': *{self.tag}*' if self.tag else '')
        result = [f'***Request:***  `{" ".join(self.tokens)}`\n',
                  f'***Rolled out:***  `{self.expr}`\n',
                  f'***Result:***\n```{self.result.describe(mode=self.game_mode)}```']
        result = ''.join(result)
        msg = '\n'.join((header, result))

        return msg

    def __str__(self):
        ret = f'<RollHander'\
              f' cmd="{self.roll}"'\
              f' from={self.ctx.author}'\
              f' guild={self.ctx.guild}'\
              f' tokens={self.tokens}'\
              f' tag="{self.tag}"'\
              f' expanded={self.expanded}'\
              f' result={self.result}>'
        return ret


class QuietRollHandler(RollHandler):

    def msg(self):
        header = f'*:game_die: {self.tag}*' if self.tag else ''
        result =f'```{self.result.describe(mode=self.game_mode)}```'
        msg = f'{header}\n{result}'

        return msg


class SekretRollHandler(RollHandler):

    def msg(self):
        header = f':game_die: from **{self.ctx.author}** in **{self.ctx.guild}**: ' + (f'*{self.tag}*' if self.tag else '')
        result = [f'***Request:***  `{" ".join(self.tokens)}`\n',
                  f'***Rolled out:***  `{self.expr}`\n',
                  f'***Result:***\n```{self.result.describe(mode=self.game_mode)}```']
        result = ''.join(result)
        return '\n'.join((header, result))


class DieResult:

    def __init__(self, expr, result):
        if not isinstance(result, int):
            raise ValueError(f'Must be scalar (got {result}).')
        self.expr = expr
        self.result = result

    def __str__(self):
        return f'{self.expr} ⤳ {self.result}'

    def __int__(self):
        return self.result


class RollList:
    def __init__(self, expr, roll):
        self.expr = expr

        if isinstance(roll, int):
            self.roll = [roll]
        else:
            self.roll = list(roll)

    def describe(self, **kwargs):
        dsc = ', '.join((str(r) for r in self.roll))
        return dsc if dsc else '0'

    def __iter__(self):
        for r in self.roll:
            yield r

    def __str__(self):
        roll = self.roll
        if len(self.roll) == 1:
            roll = self.roll[0]
        return f'{{{self.expr} ⤳ {roll}}}'

    def __repr__(self):
        return f'RollList("{self.expr}", {self.roll})'
    
    def __add__(self, other):
        if isinstance(other, RollList):
            return RollList(f'{self} + {other}', self.roll + other.roll)
        else:
            return RollList(f'{self} + {other}', (r + other for r in self.roll))

    def __neg__(self):
        return RollList(f'-{self}', (-r for r in self.roll))

    def __pos__(self):
        return RollList(f'+{self}', (+r for r in self.roll))

    def __sub__(self, other):
        return RollList(f'{self} - {other}', (r - other for r in self.roll))

    def __lt__(self, other):
        preds =  [r < other for r in self.roll]
        deltas = [abs(other - r) + int(not p) for r, p in zip(self.roll, preds)]
        return DiceDelta(self.roll, deltas, preds, f'{self} < {other}')

    def __le__(self, other):
        preds = [r <= other for r in self.roll]
        deltas = [abs(other - r) for r in self.roll]
        return DiceDelta(self.roll, deltas, preds, f'{self} <= {other}')

    def __eq__(self, other):
        return [r == other for r in self.roll]

    def __gt__(self, other):
        preds = [r > other for r in self.roll]
        deltas = [abs(other - r) + int(not p) for r, p in zip(self.roll, preds)]
        return DiceDelta(self.roll, deltas, preds, f'{self} > {other}')

    def __ge__(self, other):
        preds = [r >= other for r in self.roll]
        deltas = [abs(r - other) for r in self.roll]
        return DiceDelta(self.roll, deltas, preds, f'{self} >= {other}')


class DiceDelta:

    def __init__(self, rolls, deltas, predicates, expr=''):
        self.rolls = rolls
        self.deltas = deltas
        self.predicates = predicates
        self.expr = expr

    def describe(self, mode = None):
        if mode is not None and not isinstance(mode, GameMode):
            raise ValueError('mode must be of type GameMode')
        desc = []
        for roll, delta, pred in zip(self.rolls, self.deltas, self.predicates):
            if mode is GameMode.RT and 'd100' in self.expr:
                degrees = delta // 10
                status = SUCCESS if pred else FAILURE
                if degrees > 0:
                    result = f'{status} {degrees}°'
                else:
                    result = status
                desc.append(f'{roll:4} ⤳ {result}')
            else:
                kind = f'{SUCCESS} by' if pred else f'{FAILURE} by'
                desc.append(f'{roll:4} ⤳ {kind} {delta}')
        desc = '\n'.join(desc)
        return desc if desc else '0'


class SimpleRollConvert(commands.Converter):

    pat = r'^\d+d\d+t?$'

    async def convert(self, ctx, argument):
        match = re.match(self.pat, argument)
        log.info(f'match {argument}')

        if match is None:
            log.info(f'SimpleRollConvert: {argument} did not match')
            raise commands.BadArgument(f'{argument} is not a simple roll.')
        try:
            roll = dice.roll(argument)
        except:
            log.info(f'Failed to parse roll: {argument}')
            raise commands.BadArgument(f'Failed to parse roll: {argument}')
        else:
            log.info(f'SimpleRoll: {argument} ⤳ {roll}')
            return DieResult(argument, int(roll))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2020
# File   : zvars.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 22.02.2021

from discord.ext import commands

from .database import fetch_guild_db
from .logging import LoggingMixin


class VarCommands(commands.Cog, LoggingMixin):

    def __init__(self, bot, db):
        self.bot = bot
        self.db = db

        super().__init__()

    @commands.group(name='zvar', help='Manage variables for the server.')
    async def zvar(self, ctx):
        if ctx.invoked_subcommand is None:
            pass

    @zvar.command(name='list', help='Print current variables.')
    @fetch_guild_db
    async def zvar_list(self, ctx):
        variables = await ctx.guild_db.get_guild_vars()
        if variables:
            result = [f'**{key}**: {val}' for key, val in variables.items()]
            await ctx.send('\n'.join(result))
        else:
            await ctx.send('**No variables set.**')
    
    @zvar.command(name='set', help='Set variables for the server.')
    @fetch_guild_db
    async def zvar_set(self, ctx, var: str, val: int):
        await ctx.guild_db.set_guild_var(ctx.author.id, var, val)
        await ctx.send(f'**{var}** = {val}')

    @zvar.command(name='get', help='Print a variable value.')
    @fetch_guild_db
    async def zvar_get(self, ctx, var: str):
        val = await ctx.guild_db.get_guild_var(var)
        if val is None:
            await ctx.send(f'**{var}** is not defined.')
        else:
            await ctx.send(f'**{var}** = {val}')

    @zvar.command(name='del', help='Delete a variable.')
    @fetch_guild_db
    async def zvar_del(self, ctx, var: str):
        await ctx.guild_db.del_guild_var(var)
        await ctx.send(f'**{var}** deleted')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2020
# File   : database.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 22.02.2021

import aiofiles
import aiosql
import aiosqlite

import asyncio
from datetime import datetime
import functools
import os
import sqlite3

from .logging import LoggingMixin
from .state import GameMode
from .utils import __pkg_dir__


def load_sql_commands(*args, **kwargs):
    mode = kwargs.get('mode', 'aiosqlite')
    cmd_file = os.path.join(__pkg_dir__, 'sql', 'commands.sql')
    cmds = aiosql.from_path(cmd_file, mode)
    return cmds


class DatabaseCache(LoggingMixin):

    def __init__(self, db_dir):
        self.cache = {}
        self.db_dir = db_dir
        db_dir.mkdir(parents=True, exist_ok=True)

        super().__init__()

        self.log.info(f'DatabaseCache using dir: {db_dir}')

    async def get_guild_db(self, guild_id: int):
        self.log.info(f'Fetch database for {guild_id}')
        db = self.cache.get(guild_id, None)
        if db is None:
            self.log.info(f'Database for {guild_id} not in cache.')
            db = await ZardozDatabase.build(self.db_dir, guild_id)
            self.cache[guild_id] = db
        return db

    async def close(self):
        self.log.info(f'Closing {len(self.cache)} database connections.')
        for guild_id, db in self.cache.items():
            await db.close()


class ZardozDatabase(LoggingMixin):

    def __init__(self, con, guild_id, async_mode=True):
        self.con = con
        self.con.row_factory = sqlite3.Row
        self.guild_id = guild_id

        if async_mode:
            self.cmds = load_sql_commands()
        else:
            self.cmds = load_sql_commands(mode='sqlite3')

        self._register_aiosql_commands(self.cmds)

        super().__init__()

    def _register_aiosql_commands(self, cmds):
        for cmd_name in cmds.available_queries:
            # get the method and bind the connection to it
            cmd_func = getattr(cmds, cmd_name)
            bound = functools.partial(cmd_func, self.con)

            # build the new method for ZardozDatabase
            @functools.wraps(bound)
            def wrapped_cmd(_self, *args, **kwargs):
                return bound(*args, **kwargs)

            # add it to ZardozDatabase instance
            new_name = f'{cmd_name}_cmd'
            setattr(self, new_name, bound)

    @classmethod
    async def build(cls, db_dir, guild_id):
        spec_file = os.path.join(__pkg_dir__, 'sql', 'database.sql')
        async with aiofiles.open(spec_file, mode='r') as fp:
            spec_script = await fp.read()

        path = db_dir.joinpath(f'{guild_id}.db')
        con = await aiosqlite.connect(path)
        await con.executescript(spec_script)
        await con.commit()

        return cls(con, guild_id)

    @staticmethod
    def convert_time(timestamp):
        return datetime.fromtimestamp(timestamp).astimezone()
        
    async def close(self):
        await self.con.close()

    async def add_roll(self, member_id: int, member_nick: str, 
                             member_name: str, roll: str, tag: str, 
                             result: str):
        await self.insert_roll_cmd(member_id=member_id, member_nick=member_nick,
                                   member_name=member_name, roll=roll,
                                   tag=tag, result=result,
                                   time=datetime.now().timestamp())
        await self.con.commit()

    async def get_rolls(self, member_id=None, max_rolls=5, since=None):
        async with self.get_rolls_cursor_cmd(max_rolls=max_rolls) as cur:
            async for row in cur:
                result = dict(row)
                result['time'] = self.convert_time(result['time'])
                yield result

    async def get_last_user_roll(self, member_id: int):
        roll = await self.get_last_user_roll_cmd(member_id)
        if not roll:
            return None

        roll = dict(roll)
        roll['time'] = self.convert_time(roll['time'])
        return roll

    async def get_merged_vars(self, member_id: int):
        user_vars = await self.get_user_vars(member_id)
        guild_vars = await self.get_guild_vars()
        guild_vars.update(user_vars)
        return guild_vars

    async def set_user_var(self, member_id: int, var: str, val: int):
        await self.set_user_var_cmd(member_id=member_id, var=var, val=val)
        await self.con.commit()

    async def get_user_var(self, member_id: int, var: str):
        result = await self.get_user_var_cmd(member_id=member_id, var=var)
        return result['val'] if result else None
    
    async def get_user_vars(self, member_id: int):
        user_vars = {}
        async with self.get_user_vars_cursor_cmd(member_id=member_id) as cur:
            async for row in cur:
                user_vars[row['var']] = row['val']
        return user_vars

    async def del_user_var(self, member_id: int, var: str):
        await self.del_user_var_cmd(member_id=member_id, var=var)
        await self.con.commit()

    async def set_guild_var(self, member_id: int, var: str, val: int):
        await self.set_guild_var_cmd(member_id=member_id, var=var, val=val)
        await self.con.commit()

    async def get_guild_var(self, var: str):
        result = await self.get_guild_var_cmd(var=var)
        return result['val'] if result else None
    
    async def get_guild_vars(self):
        guild_vars = {}
        async with self.get_guild_vars_cursor_cmd() as cur:
            async for row in cur:
                guild_vars[row['var']] = row['val']
        return guild_vars

    async def del_guild_var(self, var: str):
        await self.del_guild_var_cmd(var=var)
        await self.con.commit()
    
    async def set_guild_mode(self, mode: GameMode):
        if not isinstance(mode, GameMode):
            try:
                mode = GameMode[mode]
            except KeyError:
                raise ValueError(f'{mode} is not a valid GameMode')
        await self.set_guild_var(0, 'MODE', mode)

    async def get_guild_mode(self):
        mode = await self.get_guild_var('MODE')
        if mode:
            return GameMode(mode)
        else:
            await self.set_guild_mode(GameMode.DEFAULT)
            return GameMode.DEFAULT


def fetch_guild_db(func):

    @functools.wraps(func)
    async def wrapper(self, ctx, *args, **kwargs):
        ctx.guild_db = await self.db.get_guild_db(ctx.guild.id)
        ctx.game_mode = await ctx.guild_db.get_guild_mode()
        ctx.variables = await ctx.guild_db.get_merged_vars(ctx.author.id)
        return await func(self, ctx, *args, **kwargs)

    return wrapper
